<'
# Copyright (c) 2011-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Tue Dec 13 2011
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include <cstring>
#include <unistd.h>
#include <signal.h>

#include "ros/ros.h"
#include "ros/callback_queue.h"
#include "actionlib/server/action_server.h"

#include "<"$comp">_internals.h"
#include "<"$comp">_typecopy.h"


/* <"[--- ${comp}_cntrl_task ----------------------------------------------- ]"> */

/** Control task of component <"$comp">
 *
 * This task waits for incoming messages or internal events. Upon reception
 * of a request, the corresponding callback function is invoked. Internal
 * events are handled and final replies are sent to the client.
 */

int
<"$comp">_cntrl_task(const char *instance)
{
  ros::CallbackQueue *queue = ros::getGlobalCallbackQueue();
  pthread_mutexattr_t attr;
  genom_component_data cids;
  int s, done;

  cids.context.raise = ros_server_raise;
  cids.context.raised = ros_server_raised;
  cids.context.data = &cids.context_data;

  cids.context_data.self = &cids;
  cids.context_data.ex = genom_ok;
  cids.context_data.exdetail = NULL;
  cids.context_data.exsize = 0;

  cids.node = new ros::NodeHandle(instance);

  /* disable actionlib status updates when goal status does not change */
  cids.node->setParam("actionlib_status_frequency", 0);

  s = pthread_mutexattr_init(&attr);
  s += pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
  s += pthread_mutex_init(&cids.task_lock, &attr);
  if (s) err(2, "initializing task lock");

  cids.ticks = 0;
  memset(cids.run_map, 0, sizeof(cids.run_map));
<'if {![catch {$component ids}]} {'>
  genom::ids::pinit(cids.ids);
<'}'>

  /* advertise data outports */
  genom_log_info("advertising ports");
<'foreach p [$component ports out simple] {'>
  cids.ports.<"[$p name]">.init(cids.node);
<'}'>
  s = cids.genom_state_init();
  if (s) errx(2, "error while initializing state port");

  /* create tasks */
<'foreach t [$component tasks] {'>
  cids.tasks.<"[$t name]">.context.raise = ros_server_raise;
  cids.tasks.<"[$t name]">.context.raised = ros_server_raised;
  cids.tasks.<"[$t name]">.context.data = &cids.tasks.<"[$t name]">.context_data;

  cids.tasks.<"[$t name]">.context_data.self = &cids;
  cids.tasks.<"[$t name]">.context_data.ex = genom_ok;
  cids.tasks.<"[$t name]">.context_data.exdetail = NULL;
  cids.tasks.<"[$t name]">.context_data.exsize = 0;

  cids.tasks.<"[$t name]">.spawn(&cids,
                    "<"[$t name]">", <"$comp">_<"[$t name]">_task);
<'}'>

  /* advertise services */
  genom_log_info("advertising services");
<'foreach s [$component services] {'>
<'  if {[catch {$s task} t]} {'>
  cids.services.<"[$s name]"> =
    cids.node->advertiseService(
      "<"[$s name]">", &genom_component_data::<"[$s name]">_rqstcb, &cids);
<'  } else {'>
  cids.services.<"[$s name]"> =
    new actionlib::ActionServer< genom::action::<"[$s name]"> >(
      *cids.node, "<"[$s name]">",
      boost::bind(&genom_component_data::<"[$s name]">_rqstcb, &cids, _1),
      boost::bind(&genom_component_data::<"[$s name]">_intercb, &cids, _1),
      false);
  cids.services.<"[$s name]">->start();
<'  }'>
<'}'>

<'if {![catch {$component clock-rate} rate]} {'>
  /* create timer */
  cids.clock = cids.node->createTimer(ros::Duration(<"[$rate value]">),
                                      &genom_component_data::timercb, &cids,
                                      false, true);
<'}'>

  /* 
   * If SIGUSR1 is set to SIG_IGN, it can be two things:
   *   - the parent really ignores it, so it won't care about it
   *   - the parent want to significate that it have a special handler to deal
   *   with it. In this case, it is valuable to send it. It is in particular
   *   true in the case of the daemon mode.
   */
  sig_t handler = signal(SIGUSR1, SIG_IGN);
  if (handler == SIG_IGN)
	  kill(getppid(), SIGUSR1);
  signal(SIGUSR1, handler);

  /* go */
  genom_log_info("control task initialized and running");
  done = 0;
  while(ros::ok() && !done) {
    queue->callAvailable(ros::WallDuration(1.));

    if (genom_shutdown) {
      done = 1;
      pthread_mutex_lock(&cids.task_lock);
      for(unsigned i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
        if (cids.activities.activities[i]) done = 0;
      }
      pthread_mutex_unlock(&cids.task_lock);
    }
  }

  /* cleanup */
  pthread_mutex_lock(&cids.task_lock);
<'foreach task [$component tasks] {'>
  pthread_cond_signal(&cids.tasks.<"[$task name]">.signal);
<'}'>
  pthread_mutex_unlock(&cids.task_lock);
<'foreach task [$component tasks] {'>
  pthread_join(cids.tasks.<"[$task name]">.id, NULL);
<'}'>

  genom_log_info("shutting down services");
<'foreach s [$component services] {'>
<'  if {[catch {$s task} t]} {'>
  cids.services.<"[$s name]">.shutdown();
<'  } else {'>
  delete cids.services.<"[$s name]">;
<'  }'>
<'}'>

  genom_log_info("shutting down ports");
<'foreach p [$component ports] {'>
  cids.ports.<"[$p name]">.fini();
<'}'>

  genom_log_info("shutting down control task");
<'if {![catch {$component clock-rate} rate]} {'>
  cids.clock.stop();
<'}'>
<'if {![catch {$component ids}]} {'>
  genom::ids::pfini(cids.ids);
<'}'>
  cids.node->shutdown();
  delete cids.node;

  genom_log_info("shutdown complete");
  return 0;
}


/* <"[--- genom_exec_task::spawn -------------------------------------------]"> */

void
genom_exec_task::spawn(genom_component_data *self,
                       const char *name, void *(*start)(void *))
{
  int s;

  s = pthread_cond_init(&signal, NULL);
  if (s) err(2, "initializing task condition variable");
  id = 0;

  pthread_mutex_lock(&self->task_lock);
  runnable = wakeup = false;
  status = <"$comp">_sleep;

  s = pthread_create(&id, NULL, start, self);
  if (s) err(2, "spawning task %s", name);
  do {
    s = pthread_cond_wait(&signal, &self->task_lock);
    if (s) err(2, "spawning task %s", name);
  } while(status == <"$comp">_sleep);
  pthread_mutex_unlock(&self->task_lock);
}


/* <"[--- ${comp}_cntrl_task_signal -------------------------------------]"> */

/** \brief Signal the control task that an activity is terminated
 */

static const struct genom_activity_cbd_ {
  void operator()(genom_activity *) {}
} genom_activity_cbd = {};

void
<"$comp">_cntrl_task_signal(genom_activity *a)
{
  ros::CallbackInterfacePtr cb(a, genom_activity_cbd);
  ros::getGlobalCallbackQueue()->addCallback(cb);
}


/* <"[--- genom_component_data::exec_task_signal ---------------------------]"> */

/** \brief Signal executions tasks
 */

void
genom_component_data::exec_task_signal()
{
<'foreach t [$component tasks] {'>
  if (!tasks.<"[$t name]">.runnable) {
    tasks.<"[$t name]">.runnable = true;
<'  if {[catch {$t period}]} {'>
    tasks.<"[$t name]">.wakeup = true;
    pthread_cond_signal(&tasks.<"[$t name]">.signal);
<'  }'>
  }
<'}'>
}


<'if {![catch {$task period} period]} {'>
/* --- Timer callback ------------------------------------------------------ */

void
genom_component_data::timercb(const ros::TimerEvent &e)
{
  pthread_mutex_lock(&task_lock);
  ticks++;
<'  foreach task [$component tasks] {'>
<'
    set rate [[$component clock-rate] value]
    set ticks [expr {int([$period value]/$rate)}]
    if {![catch {$task delay} delay]} {
      set d [expr {int([$delay value]/$rate)}]
    } else {
      set d 0
    }
'>
  if (tasks.<"[$task name]">.runnable &&
      ticks % <"$ticks"> == <"$d">) {
    tasks.<"[$task name]">.wakeup = true;
    pthread_cond_signal(&tasks.<"[$task name]">.signal);
  }
<'  }'>
  pthread_mutex_unlock(&task_lock);
}
<'}'>


<'foreach s [$component services] {'>


<'  if {[catch {$s task} task]} {'>
/* <"[--- Control service [$s name] ----------------------------------------]"> */

bool
genom_component_data::<"[$s name]">_rqstcb(
  genom::srv::<"[$s name]">::input &in, genom::srv::<"[$s name]">::output &out)
{
  genom::srv::<"[$s name]">::locals locals;
  pthread_mutex_lock(&task_lock);
  genom_event s;

  genom_log_debug("handling service %s", "<"[$s name]">");

  s = <"[$s name]">_controlcb(locals, in, out);
  out.genom_success = s == genom_ok;
  if (!out.genom_success)
    genom_<"$comp">_encodex(out.genom_exdetail, s,
                   context.raised(NULL, &context));

  if (out.genom_success) {
    run_map[<"$COMP">_<"[$s name]">_RQSTID] = true;

<'    if {[llength [$s interrupt]]} {'>
    /* interrupt incompatible activities */
    for(size_t i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
      genom_activity *c = activities.activities[i];
      if (!c) continue;
      switch(c->rqstid) {
<'      foreach c [$s interrupt] {'>
        case <"$COMP">_<"[$c name]">_RQSTID:
<'      }'>
          c->stop("<"[$s name]">");
          break;

        default: break;
      }
    }
<'    }'>
  }

  exec_task_signal(); /* wake up sleeping activities */

  genom_log_debug("done service %s", "<"[$s name]">");
  pthread_mutex_unlock(&task_lock);
  return true;
}
<'  } else {'>
/* <"[--- Execution service [$s name] --------------------------------------]"> */

template<> int
genom_activity_data_<"[$s name]">::interrupt()
{
  int wait = 0;
<'    if {[llength [$s interrupt]]} {'>
  genom_activity *c;
  size_t i;

  /* interrupt incompatible activities */
  for(i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
    c = self->data->self->activities.activities[i];
    if (!c || c->id == id) continue;
    switch(c->rqstid) {
<'      foreach c [$s interrupt] {'>
      case <"$COMP">_<"[$c name]">_RQSTID:
<'      }'>
        if (c->status == ACT_STOP ||
            c->status == ACT_RUNNING || c->status == ACT_STOPPING)
          wait = 1;
        c->stop("<"[$s name]">");
        break;

      default: break;
    }
  }
<'    }'>

  return wait;
}

void
genom_component_data::<"[$s name]">_rqstcb(
  actionlib::ServerGoalHandle< genom::action::<"[$s name]"> > &rqst)
{
  genom_activity_data_<"[$s name]"> *a;
  genom_event s;

  genom_log_debug("handling service %s", "<"[$s name]">");

  a = new genom_activity_data_<"[$s name]">(
    <"$COMP">_<"[$s name]">_RQSTID, tasks.<"[$task name]">.id,
    rqst, &context, &tasks.<"[$task name]">.context);

  pthread_mutex_lock(&task_lock);
  s = activities.alloc(a);
  if (!s)
    s = <"[$s name]">_controlcb(a->locals, a->in, a->out);
  if (s) {
    a->reject();
    activities.free(a);
    pthread_mutex_unlock(&task_lock);
    delete a;
    return;
  }

  a->accept();

  int delay = a->interrupt();

  /* update state port */
  genom_state_component *state = ports.genom_state.handle.data(&context);
  if (state->activity._length < state->activity._maximum) {
    a->stateid = state->activity._length++;
    state->activity._buffer[a->stateid].id = a->id;
    strncpy(state->activity._buffer[a->stateid].name, "<"[$s name]">",
            sizeof(state->activity._buffer[a->stateid].name));
    if (delay)
      state->activity._buffer[a->stateid].state[0] = *"";
    else
      strncpy(state->activity._buffer[a->stateid].state, <"$comp">_start,
              sizeof(state->activity._buffer[a->stateid].state));
    ports.genom_state.publish();
  } else
    a->stateid = -1;

  if (delay) {
    /* wait for other activities */
    pthread_mutex_unlock(&task_lock);
    return;
  }

  /* start activity */
  a->run();
  exec_task_signal();

  genom_log_debug("done service %s", "<"[$s name]">");
  pthread_mutex_unlock(&task_lock);
}

void
genom_component_data::<"[$s name]">_intercb(
  actionlib::ServerGoalHandle< genom::action::<"[$s name]"> > &rqst)
{
  actionlib_msgs::GoalID gid;
  genom_activity *a;

  pthread_mutex_lock(&task_lock);
  gid = rqst.getGoalID();
  a = activities.bygid(gid);
  if (a == NULL) {
    pthread_mutex_unlock(&task_lock);
    genom_log_warn(
      "cancel request for non-existent activity %s", gid.id.c_str());
    return;
  }

  a->stop("<"[$s name]">");

  exec_task_signal();
  pthread_mutex_unlock(&task_lock);
}

<'  }'>
<'}'>


/* --- Activity callback --------------------------------------------------- */

void
genom_component_data::cntrl_task_activitycb(genom_activity *a)
{
  genom_state_component *state = ports.genom_state.handle.data(&context);
  genom_activity *c;
  size_t i;
  int sid;

  pthread_mutex_lock(&task_lock);
  if (a->status == ACT_ETHER && a->codel == <"$comp">_ether)
    run_map[a->rqstid] = true;
  a->report();
  sid = a->stateid;
  activities.free(a);
  exec_task_signal();

  /* update state port */
  if (sid >= 0) {
    state->activity._length--;
    memcpy(&state->activity._buffer[sid],
           &state->activity._buffer[sid+1],
           (state->activity._length - sid) *
           sizeof(state->activity._buffer[sid]));
  }

  /* check for pending activities */
  for(i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
    c = activities.activities[i];
    if (!c) continue;
    if (sid >= 0 && c->stateid > sid) c->stateid--;
    if (c->status != ACT_START) continue;
    if (c->interrupt()) continue;

    /* update state port */
    if (c->stateid >= 0) {
      strncpy(state->activity._buffer[c->stateid].state, <"$comp">_start,
              sizeof(state->activity._buffer[c->stateid].state));
    }

    c->run();
  }

  ports.genom_state.publish();
  pthread_mutex_unlock(&task_lock);
  delete a;
}
